# Copyright 2023 Pranjal Choudhury (p.choudhury@iitg.ac.in)
# 
# Permission is hereby granted, free of charge, to any person obtaining 
# a copy of this software and associated documentation files (the "Software"), 
# to deal in the Software without restriction, including without limitation 
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS IN THE SOFTWARE.


######################################################################################################################################################################################################################################################################
# importing essential python libraries
######################################################################################################################################################################################################################################################################

import os
import cv2
import numpy as np
import pandas as pd
import skimage.io as io
import matplotlib.pyplot as plt
from skimage.transform import resize

import utilities.PeakLocalMaxima
import utilities.CentreOfMass
import utilities.BlobDetection
import utilities.OneD_LSfitting
import utilities.TwoD_LSfitting
import utilities.TwoD_MLE
import utilities.ImageVisualization
import utilities.RemoveIncorrectLocalizations

def main(input, init_params, threshold, PSF_radius, window_size, neighbourhood_size, reconstruction_type, crude_localization_type, sub_pixel_localization_method):
    
    # Create results directory
    dir_results = './logs'
    if not os.path.exists(dir_results):
        os.mkdir(dir_results)
    
    try:
        init_params = init_params.split(",")
        init_params = [float(num) for num in init_params]
    except AttributeError:
        raise AttributeError('Provide a valid init_params list. Numbers should be separated by commas')
    
    # Load images
    image_stack=io.imread(input)                                                                 # reading the image                                                    
    number_of_images, image_dim_1, image_dim_2=image_stack.shape                                            # extract the array dimensions. i.e. n, dim1. dim2
    image_shape=image_dim_1, image_dim_2                                                                    # dimensions of each image frame

    #initialize array for real time scatter plot display of reconstructed image
    sr_im = np.zeros((image_shape[0]*10, image_shape[1]*10))

    ######################################################################################################################################################################################################################################################################
    # define localization parameters
    ######################################################################################################################################################################################################################################################################

    ##  various lists that will be filled in the subsequent loop 
    crude_localized_molecule_x = []
    crude_localized_molecule_y = []

    molecule_intensity = []
    localized_molecule_x = []
    localized_molecule_y = []
    sigma_x = []
    sigma_y = []

    # initializing image count 
    image_count = 0

    # initializing the loop
    # for image_no in range(0,number_of_images):
    while image_count < number_of_images:

        #####################################################################################################################################################################################################################
        # loading of stack of image directly and calling and localizing them 
        #calling individual files from the stack and localizing them
        image = image_stack[image_count]/image_stack[image_count].max()*255.
        image_count+=1
                
        #####################################################################################################################################################################################################################
        # image is now loaded to memory, the following code will localize the PSFs
        
        ######################################################################################################################################################################################
        #crude localizatiom of PSF;  Input: {image, threshold, , PSF radius / neighbourhood size}  Output: {crude coordinates, method}
        ######################################################################################################################################################################################
        # detection using peak local maxima
        if crude_localization_type == 'crude_plm':
            method = 'PeakLocalMaxima'                       
            crude_coordinates = utilities.PeakLocalMaxima.Detection(image=image, threshold=threshold, PSF_size=PSF_radius)
            
        # detection using Centre of Mass
        if crude_localization_type == 'crude_com':
            method = 'COM'                       
            crude_coordinates = utilities.CentreOfMass.Detection(image=image,threshold=threshold,neighborhood_size=neighbourhood_size)
        
        # detection using Blob LoG
        if crude_localization_type == 'crude_bl':
            method = 'BlobDetection'                      
            crude_coordinates = utilities.BlobDetection.Detection(image=image, threshold=threshold, PSF_size=PSF_radius*2)

        ### if reconstruction is based on crude localization only, the following code will run ###
        if reconstruction_type == 'rec_crude':
            
            for ii, jj in zip(crude_coordinates[0][0], crude_coordinates[1][0]):
                sr_im[int(jj * 10 ), int(ii * 10)] = 1
            if image_count % 100==0:		# can be commented out to save time if real time display id not needed
                print(image_count)			# can be commented out to save time if real time display id not needed
                # Display the output image 
                #cv2.imshow('output', sr_im)		# can be commented out to save time if real time display id not needed
                #cv2.waitKey(1)			# can be commented out to save time if real time display id not needed

            crude_localized_molecule_x.append(crude_coordinates[0][0])
            crude_localized_molecule_y.append(crude_coordinates[1][0])

        ### if reconstruction is based on crude localization followed by sub pixel localization, the following code will run ###

        if reconstruction_type == 'rec_sub':
        ######################################################################################################################################################################################
        # sub pixel localization ;  Input: {crude coordinates, image, window size, initial guess, method}  Output: {params (PSF intensity, x_position, y_position, sigma_x, sigma_y)}
        ######################################################################################################################################################################################
            #sub pixel localization using 1D Least Square fitting
            if sub_pixel_localization_method == 'sub_1d': params = utilities.OneD_LSfitting.localizer(coordinates=np.array(crude_coordinates), images=image, window_size=window_size, initial_guess=init_params,method=method)

            #sub pixel localization using 2D Least Square fitting
            if sub_pixel_localization_method == 'sub_2d': params = utilities.TwoD_LSfitting.localizer(coordinates=np.array(crude_coordinates), images=image, window_size=window_size, initial_guess=init_params,method=method)
                
            #sub pixel localization using 2D MLE
            if sub_pixel_localization_method == 'sub_mle': params = utilities.TwoD_MLE.localizer(coordinates=np.array(crude_coordinates), images=image, window_size=window_size, initial_guess=init_params,method=method)

            # construct a binary image using the localizations for dynamic display 
            for ii, jj in zip(params[1], params[2]):
                if ((ii>=0 and ii <= image_shape[0]) and (jj>=0 and jj <= image_shape[1])):
                    sr_im[int(jj * 10), int(ii * 10)] = 1		

            #localization parameters: 
            molecule_intensity.append(params[0])            # amplitude of the detected PSFs
            localized_molecule_x.append(params[1])          # x cordinate of the detected PSFs
            localized_molecule_y.append(params[2])          # y cordinate of the detected PSFs
            sigma_x.append(params[3])                       # sigma_x of the fitted Gaussian of the PSFs
            sigma_y.append(params[4])                       # sigma_x of the fitted Gaussian of the PSFs
        #####################################################################################################################################################################################################################
            
    ######################################################################################################################################################################################################################################################################
    ## conctaenating the lists into 1D arrays for visualization and removing incorrect localizations
    ######################################################################################################################################################################################################################################################################
    if reconstruction_type == 'rec_sub':
        detection_x = np.concatenate(localized_molecule_x,axis=0)
        detection_y = np.concatenate(localized_molecule_y,axis=0)

    else:
        detection_x = np.concatenate(crude_localized_molecule_x,axis=0)
        detection_y = np.concatenate(crude_localized_molecule_y,axis=0)

    # removing incorrect localizaations 
    detection_x, detection_y = utilities.RemoveIncorrectLocalizations.remove_outliers(detection_x,detection_y, image_shape)

    ######################################################################################################################################################################################################################################################################
    # saving the detections to csv file 
    ######################################################################################################################################################################################################################################################################
    df=pd.DataFrame({'detection_x': detection_x, 'detection_y': detection_y})
    df.to_csv(dir_results+f'/detections.csv',index=False)
    print(df)

    ######################################################################################################################################################################################################################################################################
    ## constructiong the super resolved image
    ######################################################################################################################################################################################################################################################################
    image_sc = utilities.ImageVisualization.scatter_plot(image_shape,detection_x,detection_y, mag=10)
    image_hs = utilities.ImageVisualization.histogram(image_shape,detection_x,detection_y, mag=10)
    image_av = utilities.ImageVisualization.average_shifted_histogram(image_shape,detection_x,detection_y, mag=10)
    
    ######################################################################################################################################################################################################################################################################
    #visualizing and saving the super-resolved images
    ######################################################################################################################################################################################################################################################################
    cv2.imwrite(dir_results+'/input.png', np.mean(image_stack, axis=0))
    plt.imsave(fname=dir_results+'/scatter_plot.png', arr=image_sc,cmap='gray')
    plt.imsave(fname=dir_results+'/histogram.png', arr=image_hs,cmap='gray',vmin=0,vmax=2)
    plt.imsave(fname=dir_results+'/ASH.png', arr=image_av,cmap='gray', vmax=2)

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--input", type=str, required=True)
    parser.add_argument("--init_params", type=str, required=True)
    parser.add_argument("--threshold", type=int, required=True)
    parser.add_argument("--psf_radius", type=int, required=True)
    parser.add_argument("--window_size", type=int, required=True)
    parser.add_argument("--neighbourhood_size", type=int, required=True)
    parser.add_argument("--reconstruction_type", type=str, required=True)
    parser.add_argument("--crude_localization_type", type=str, required=True)
    parser.add_argument("--sub_pixel_localization_type", type=str, required=True)

    args = parser.parse_args()
    main(args.input, args.init_params, args.threshold, args.psf_radius, args.window_size, args.neighbourhood_size, args.reconstruction_type, args.crude_localization_type, args.sub_pixel_localization_type)